/* $Id: parser.mly,v 1.11 2001/11/12 09:09:25 malekith Exp $ 
 *
 * Parser in ocamlyacc for the gont language.
 */
%{
open Gont
;;

let locus () = Lexer.get_current_location ()

let make_exp raw =
  {raw_exp = raw;
   exp_ty = None;
   exp_loc = locus ()}

let bin op e1 e2 =
  make_exp (SimpleBinOp(op, e1, e2))
  
let make_stmt raw =
  {raw_stmt = raw;
   stmt_loc = locus ()}
;;

let unary op e =
  {raw_exp = SimpleUnOp(op, e);
   exp_ty = None;
   exp_loc = locus ()}
;;

%}

%token <string> INT
%token <string> FLOAT
%token <string> ID
%token <string> STRING
%token KW_INT KW_RETURN KW_VOID KW_SKIP KW_BOOL KW_FLOAT KW_PUBLIC 
%token KW_EXTERN_C KW_STRING
%token ARROW LSQUARE RSQUARE
%token PLUS MINUS TIMES DIV EQUAL
%token LPAREN RPAREN LBRACE RBRACE SEMICOLON COMMA

/* lowest precedence */
%right EQUAL
%left PLUS MINUS
%left TIMES DIV
%nonassoc UMINUS
%nonassoc APPLICATION
/* highest precedence */

%start decl             /* the entry point */

%type <Gont.decl> decl

/* can be converted to Ksi type (using Gont.ksi_type) */
%type <Gont.ty> type_exp

%%

decl:
          func          { $1 }
        | global_var    { $1 }
        | extern_c      { $1 }
;


id: 
        ID              { $1 }
;

storage:
          KW_PUBLIC       { (locus(), Public) }
        | /* empty */     { (locus(), Local) }
;

func:
        storage type_exp id LPAREN func_args RPAREN block
                { let (l, s) = $1 in
                  FuncDecl {fn_storage = s;
                            fn_loc = l;
                            fn_ret_type = $2; 
                            fn_name = $3;
                            fn_parms = $5;
                            fn_body = $7} }
;

global_var:
       storage type_exp id SEMICOLON
                { let (l, s) = $1 in GlobalDecl(s, $2, $3, l) }
;

/* from now on ne_ == non_empty_ */

ne_func_args:
          func_arg                      { [$1] }
        | func_arg COMMA ne_func_args   { $1 :: $3 }
;

func_args:
          /* empty */    { [] }
        | ne_func_args   { $1 }
;

func_arg:
        type_exp id             { ($1, $2) }
;

proto_arg:
          type_exp              { $1 }
        | type_exp id           { $1 }
;

ne_proto_args:
          proto_arg                             { [$1] }
        | proto_arg COMMA ne_proto_args         { $1 :: $3 }
;

proto_args:
          /* empty */           { [] }
        | ne_proto_args         { $1 }
;

proto:
        type_exp id LPAREN proto_args RPAREN
                { ($1, $2, $4) }
;

extern_c:
        KW_EXTERN_C proto SEMICOLON
        {
          let (t, n, a) = $2 in
            ExternC {xc_ret_type = t;
                     xc_name = n;
                     xc_parms = a;
                     xc_loc = locus ()}
        }
;

block:
        LBRACE stmts RBRACE     
        { 
          make_stmt (match $2 with
                        [] -> Skip
                      | _ -> Block $2) 
        }
;

stmts:
          /* empty */      { [] }
        | stmt stmts       { $1::$2 }
;

stmt:
          normal_stmt SEMICOLON         { $1 }
        | block                         { $1 }
;

normal_stmt:
          expr                  { make_stmt (Exp $1) }
        | KW_RETURN             { make_stmt ReturnVoid }
        | KW_RETURN expr        { make_stmt (Return $2) }
        | KW_SKIP               { make_stmt Skip }
;

variable:
        id                      { make_exp (Var $1) }
;

primary:
          INT                     { make_exp (IntConst $1) }
        | FLOAT                   { make_exp (FloatConst $1) }
        | STRING                  { make_exp (StringConst $1) }
        | variable                { $1 }
	| LPAREN expr RPAREN      { $2 }
;

expr:
          primary                 { $1 }
        | call %prec APPLICATION  { $1 }
	| expr PLUS expr          { bin Ksi.Plus $1 $3 }
	| expr MINUS expr         { bin Ksi.Minus $1 $3 }
	| expr TIMES expr         { bin Ksi.Mult $1 $3 }
	| expr DIV expr           { bin Ksi.TruncDiv $1 $3 }
	| MINUS expr %prec UMINUS { unary Ksi.Negate $2 }
;

ne_expr_list:
          expr                           { [$1] }
        | expr COMMA ne_expr_list        { $1 :: $3 }
;

expr_list:
          /* empty */    { [] }
        | ne_expr_list   { $1 }
;


call:
        primary LPAREN expr_list RPAREN  { make_exp (Call($1, $3)) }
;
        
simple_type:
	  KW_VOID	{ Void }
	| KW_INT	{ Int }
        | KW_BOOL       { Bool }
        | KW_FLOAT      { Float }
        | KW_STRING     { String }
;

/* Function type is written as follows:
 * [ t1 n1, t2 n2, ..., tn nn ] -> t, ni entries can be ommited, n >= 0 */
function_type:
        LSQUARE proto_args RSQUARE ARROW type_exp       { Func($5, $2) }
;

type_exp: 
	  simple_type           { $1 }
        | function_type         { $1 }
;
